use32
;//cpu 686

;// DLL - compile into DLL ?
%define DLL

;// Debug version can auto-unload:
%define Debug

;//%define pe_ImageBase 0F910000h
%define pe_ImageBase 0EFE0000h

end_Code equ 800h  ;// update in 200h increments...

%ifdef DLL
%include "MozGcGc_dllHeader.inc"
%endif ;DLL

seg_Code equ pe_ImageBase + 1000h
start_Code equ $

;//----------------------------------------------
;// DLL entry point at 1000h at start of code specified in dllHeader
%ifdef DLL
jmp dll_Entry
%endif ;DLL

;// see InternalPerformThread, PerformCleanup below...

%ifdef Debug
  cWaitSleep equ 15000
%else ;Normal
  cWaitSleep equ 60000  ;// minute by default
%endif ;Normal

;// here could be pointer to GetModuleHandleA from imports, if the DLL has any...
imp_GetModuleHandle equ (seg_Code+ ($-start_Code))
imp__GetModuleHandle:
dd 0

;//----------------------------------------------
;// Types:

;// type TKernel32 - kernel32.dll functions used:
  k32_Handle                      equ 0
  k32_GetModuleHandle             equ 4 ;// for GetProcAddress
  ;//k32_VirtualAlloc                equ 8
  k32_VirtualFree                 equ 8
  k32_VirtualQuery                equ 12
  k32_CloseHandle                 equ 16
  k32_CreateThread                equ 20
  k32_Sleep                       equ 24
%ifdef Debug
  k32_FreeLibraryAndExitThread    equ 28
%endif ;Debug
  ;//
  k32__sizeOf                     equ 32

%ifndef Debug
  n_Kernel_Funcs equ 6  ;// function names in InitKernel
%else ;Debug
  n_Kernel_Funcs equ 7  ;// function names in InitKernel
%endif ;Debug

;//----------------------------------------------
;// BSS data segment:

  bss_Started           equ 0    ;// Boolean byte
  bss_idThread          equ 4    ;// dword
  bss_ThreadStack       equ 8    ;// Pointer on top of bg-thread stack, for restarts...
  
%ifdef Debug
  bss_Stop              equ 12   ;// byte - set to non-zero to stop and unload...
%endif ;Debug
  
  bss_NCleanup          equ 10h  ;// dword - count of PerformCleanup runs (i.e. also count of minutes waiting)
  bss_NReleased         equ 14h  ;// dword - count of 1Mb blocks released...
  bss_NRestarts         equ 18h  ;// dword - count of restarts...
  
  bss_Test              equ 20h
  
  bss_Kernel32          equ 50h  ;// g_Kernel record defined k32_* above...
  
%ifdef Debug
  bss_LogWrite          equ 7Ch
  bss_LogBuffer         equ 80h
  size_LogBuffer        equ 2048
%endif ;Debug

;// some constants:

  MEM_DECOMMIT equ 4000h
  MEM_RELEASE  equ 8000h
  MEM_RESERVE  equ 2000h
  PAGE_NOACCESS  equ 1
  PAGE_READWRITE equ 4
  PAGE_EXECUTE_READWRITE equ 40h
  MEM_COMMIT equ 1000h

;// type TMemoryBasicInformation
  mbi_Address      equ 0
  mbi_AllocBase    equ 4
  mbi_AllocProtect equ 8
  mbi_RegionSize   equ 12
  mbi_State        equ 16
  mbi_Protect      equ 20
  mbi_Type9        equ 24
  mbi__sizeOf      equ 28

;//----------------------------------------------
;// function getDataSeg(): Pointer - returns pointer to BSS segment of this dll...
;// avoids relocatable pointers in code segment, can run from any address...
getDataSeg:
  call .here
  .here:
  pop  eax
  ;//and  eax,-10000h  ;// ImageBase
  ;//mov  ah,rva_Bss>>8 ;//add  eax,rva_Bss  ;// rva_Bss = 8000h defined in *_dllHeader.inc
  mov  ax,rva_Bss
  ret

;//-----------------------------------------------------------------

;// function dll_Entry(a,b,c): int(1) stdcall
dll_Entry:
  ;//
  call getDataSeg
  lea  edx,[eax+bss_Started]
  cmp  byte [edx],0
  jne  .already
  inc  byte [edx]
  ;// Start bg-thread now:
  call get_Kernel
  call StartBgThread
  ;//
.already:
  push byte 1
  pop  eax
  retn 12


;// function StartBgThread(g_Kernel)
StartBgThread:
  push ebp
  mov  ebp,esp
  push eax  ;// [ebp-4] = g_Kernel
  ;// H:=CreateThread(nil,0,func,arg,0,^idThread)
  call getDataSeg
  add  eax,byte bss_idThread
  push eax
  xor  edx,edx
  push edx
  mov  ecx,[ebp-4]
  push ecx
  call .here2
  .here2:
  add  dword [esp],byte (BgThreadProc - .here2)
  push edx
  push edx
  call dword [ecx+k32_CreateThread]
  test eax,eax
  jz   .failed
  ;// CloseHandle(H)
  mov  edx,[ebp-4]
  push eax
  call [edx+k32_CloseHandle]
  ;//
.ok:
.failed:
  mov  esp,ebp
  pop  ebp
  ret

;// function BgThreadProc(g_Kernel): Longint; stdcall;
;// [esp+4]=[ebp+8]=g_Kernel
BgThreadProc:
  push ebp
  mov  ebp,esp
  push ebx ;// save registers even if we do not modify them at this level, because of possible dirty restarts...
  push edi
  push esi
  ;//
  ;// some stack reserve:
  push byte 40h
  pop  ecx
  xor  eax,eax
.clcy:
  push eax
  dec  ecx
  jnz  .clcy
  ;//
  push ebp
  call getDataSeg
  mov  [eax+bss_ThreadStack],esp
  ;//
  mov  eax,[ebp+8]
  call PerformThread
  ;//
  lea  esp,[ebp-12]
  pop  esi
  pop  edi
  pop  ebx
  pop  ebp
  retn 4

;// function RestartThread() -- restarts bg-thread in case of exception... does not return to caller...
RestartThread:
  call getDataSeg
  inc  dword [eax+bss_NRestarts]
  mov  edx,[eax+bss_ThreadStack]
  ;//
  xor  eax,eax
  db 64h,8Bh,08h ;// mov  ecx,fs:[eax]
  mov  ecx,[ecx]
  db 64h,89h,08h ;// mov  fs:[eax],ecx
  ;//
  mov  esp,edx
  mov  ebp,[edx]
  ;//
  push byte 127
  call Sleep
  ;//
  mov  eax,[ebp+8] ;// g_Kernel
  ;//jmp short PerformThread -- is just below...

;// function PerformThread(g_Kernel) -- usually does not return...
PerformThread:
  push ebp
  mov  ebp,esp
  ;//
  call .pushed_EHandler
.EHandler:
  jmp short RestartThread
.pushed_EHandler:
  ;//
  push eax ;// Param = g_Kernel
  ;//
  call .pushed_Body
.ThreadBody:
  jmp short InternalPerformThread
.pushed_Body:
  ;//
  call ZeroTry
  ;//
  mov  esp,ebp
  pop  ebp
  ret

;// function InternalPerformThread(g_Kernel) -- usually does not return...
InternalPerformThread:
  push ebp
  mov  ebp,esp
  call get_Kernel
  push eax      ;// [ebp-4] = g_Kernel
  push byte 0   ;// [ebp-8] = nFailed = 0
  ;//
.cycle:
  call PerformCleanup
  xor  edx,edx  ;// reset nFailed counter if ok...
  test al,al
  jnz  .ok
  ;//
  mov  edx,[ebp-8]
  inc  edx
  cmp  edx,byte 8
  jae  .break
.ok:
  mov  [ebp-8],edx
  ;//
.wait:
  push cWaitSleep    ;// one minute by default
  call Sleep
  ;//
  ;// note it takes a minute in Sleep above before this byte is checked...
%ifdef Debug
  call getDataSeg
  cmp  byte [eax+bss_Stop],0
  jne  .dbg_stop
%endif ;Debug
  ;//
  jmp  short .cycle
  ;//
.break:
  ;// finish???
.Finish:
  call getDataSeg
  ;//mov  [eax+bss_Test],esp
  and  dword [eax+bss_idThread],byte 0
  ;//
  xor  eax,eax
  mov  esp,ebp
  pop  ebp
  ret
%ifdef Debug
.dbg_stop:
  ;// FreeLibraryAndExitThread(hInstance,0)
  mov  eax,[ebp-4] ;//call get_Kernel
  mov  edx,eax
  xor  dx,dx ;//and  edx,-10000h
  push byte 0
  push edx
  call dword [eax+k32_FreeLibraryAndExitThread] ;// call does not return...
  db 0CCh
%endif ;Debug

;//----------------------------------------------------------------
;// function PerformCleanup(): Boolean byte
PerformCleanup:
  push ebp
  mov  ebp,esp
  push edi
  push esi
  push ebx
  ;//
  xor  eax,eax
  push eax      ;// [ebp-16] = Result Boolean
  push eax      ;// [ebp-20] = bufSize
  push eax      ;// [ebp-24] = first_scan
  mov  eax,esp
  push eax      ;// [ebp-28] = stack_top (different in second scan)
  ;//
  call getDataSeg
  inc  dword [eax+bss_NCleanup]
  ;//
  push byte 120
  pop  ebx
  jmp  short .find_abandoned_chunks
.buf_small:
.second_scan:
  mov  esp,[ebp-28] ;// stack_top
  add  ebx,byte 64
  cmp  ebx,[ebp-20] ;// check sizeNeeded
  ja   .find_abandoned_chunks
  mov  ebx,[ebp-20]
.find_abandoned_chunks:
  add  ebx,byte 16
  mov  [ebp-20],ebx
  sub  esp,ebx
  xor  eax,eax
  push eax
  push eax ;// touch stack there, init counts to 0...
  mov  eax,esp
  mov  edi,eax
  lea  edx,[ebp-20]
  call GetAbandonList
  test eax,eax
  jl   .buf_small
  ;// if(!Count_of_JSRuntime_found)return false;
  cmp  dword [edi],byte 0
  jle  .fail
  ;// if(!Count_of_AbandonedChunks)return true;
  mov  byte [ebp-16],1 ;// Result:=true -- some JSRuntime found...
  mov  ecx,[edi+4] ;// Count_of_AbandonedChunks
  add  edi,byte 8 ;// point to dwList
  test ecx,ecx
  jz   .ok
  ;//
  xor  eax,eax
  cmp  dword [ebp-24],eax ;// first_scan
  jnz  .done_second
  mov  [ebp-24],edi
  ;//
  mov  eax,edi
  mov  edx,ecx
  call MarkAbandonChunks
  ;//
  ;// there are some abandoned chunks, first scan done, wait and scan again...
  push 15000
  call Sleep
  ;//
  xor  eax,eax
  push eax
  push eax
  mov  [ebp-28],esp ;// new stack_top -- do not overwrite first scan by second one...
  jmp  short .second_scan
  ;//
.done_second: ;// here edi=second_scan dwList, ecx=count_second, [ebp-24]=first_scan, at offset -4 count
  ;//
  ;// for all in second_scan, which were also in first scan:
  ;// all in first scan were marked...
  test ecx,ecx
  mov  ebx,ecx ;// count_second
  mov  esi,edi ;// list_second
  jle  .break2
.cy2:
  lodsd ;// mov eax,[esi] + add esi,byte 4
  mov  edi,eax
  mov  edx,[ebp-24]
  call DwIndexOf
  inc  eax
  jz   .skip2
  mov  eax,edi
  call IsChunkMarked
  test al,al
  jz   .skip2
  ;// release abandoned chunk
  ;//mov  eax,edi
  call ReleaseChunk_edi
  ;//
.skip2:
  dec  ebx
  jnz  .cy2
.break2:
.fail:
.ok:
  ;//
  lea  esp,[ebp-16]
  pop  eax ;// Result
  pop  esi
  pop  edi
  pop  ebx
  pop  ebp
  ret

;// function DwIndexOf(value,List): integer;  // returns -1 if not found, List[-1] is Count (at offset -4)
DwIndexOf:
  push  edi
  mov   edi,edx
  mov   ecx,[edx-4]
  test  ecx,ecx
  jle   .not_found
  repne scasd
  jne   .not_found
  neg   edx
  lea   eax,[edi+edx-4]
  shr   eax,2
  pop   edi
  ret
.not_found:
  or    eax,byte -1
  pop   edi
  ret

;// function MarkAbandonChunks(List,Count) -- mark all in Reserved field to detect released and reused...
MarkAbandonChunks:
  push edi
  push esi
  mov  edi,eax
  test edx,edx
  mov  esi,edx
  jle  .break
.cy:
  mov  eax,[edi]
  call MmIsBadWrite
  test eax,eax
  jnz  .skip
  ;//
  mov  eax,[edi]
  call IsAbandonedChunk
  cmp  al,1
  jne  .skip
  ;//
  mov  ecx,[edi]
  add  edi,byte 4
  xor  eax,eax
  mov  edx,'fre' ;// 0E9657266h
  lock cmpxchg [ecx+0fffcch],edx ;// write if there is 0
.skip:
  dec  esi
  jnz  .cy
.break:
  pop  esi
  pop  edi
  ret

;// function IsChunkMarked(ptr): Boolean
IsChunkMarked:
  lea  eax,[eax+0fffcch]
  push eax
  call MmIsBadWrite
  test eax,eax
  pop  eax
  jnz  .bad
  cmp  dword [eax],'fre'
.bad:
  setz al
  ret

;// function ReleaseChunk_edi(edi=Chunk) -- in Chunk in edi ...
ReleaseChunk_edi:
  ;//push edi
  ;//mov  edi,eax
  mov  eax,edi
  call IsAbandonedChunk
  cmp  al,1
  jne  .diff
  ;//
  %ifdef Debug
  ;//mov  eax,edi
  call LogReleased_edi
  %endif ;Debug
  ;//
  push MEM_RELEASE
  push byte 0
  push edi
  call VirtualFree
  ;//
  test eax,eax
  jz   .diff
  ;//
  call getDataSeg
  inc  dword [eax+bss_NReleased]
  ;//
.diff:
  ;//pop  edi
  ret

%ifdef Debug
;// function LogReleased_edi(edi=Chunk)
LogReleased_edi:
  call getDataSeg
  add  eax,byte bss_LogWrite
  mov  edx,[eax]
  add  edx,byte 4
  cmp  edx,size_LogBuffer
  jbe  .write_here
.write_restart:
  xor  edx,edx
.write_here:
  mov  [eax],edx
  mov  [eax+edx+bss_LogBuffer-bss_LogWrite-4],edi
  ret
%endif ;Debug

;//----------------------------------------------------------------
;// function GetAbandonList(outList,var BufSize): integer -- get list of abandoned 1Mb blocks...
;// BufSize is in/out, returns -1 on overflow error and updates BufSize with actual bytes needed/returned
;// dwords in outList:
;// [0]=Count_of_JSRuntime_found
;// [1]=Count_of_AbandonedChunks
;// [2..2+Count]=AbandonedChunks
GetAbandonList:
  push ebp
  mov  ebp,esp
  push eax         ;// [ebp-4] = outList
  push edx         ;// [ebp-8] = ^outSize
  push dword [edx] ;// [ebp-12] = outSize available
  xor  edx,edx
  push edx         ;// [ebp-16] = spare
  push edx         ;// [ebp-20] = Count_of_JSRuntime_found
  add  edx,byte 8
  push edx         ;// [ebp-24] = size needed := 8
  push byte 4      ;// [ebp-28] = dwSkip
  push eax         ;// [ebp-32] = BufWrite (update later)
  sub  esp,byte 32 ;// [esp] = temp buffer - used also as [edx-48] below
  ;//
  xor  edx,edx
  test eax,eax
  jnz  .buf_ok
  mov  [ebp-12],eax ;// if outList missing, outSize available is 0...
  mov  [ebp-28],eax ;// and dwSkip=0 also...
.buf_ok:
  cmp  dword [ebp-12],byte 8
  jg   .size_ok
  ;//xor  edx,edx
  mov  [ebp-12],edx
  mov  [ebp-4],edx
  mov  [ebp-28],edx
.size_ok:
  ;//
  mov  eax,[ebp-4] ;// OutList
  test eax,eax
  jnz  .init_out
  mov  eax,esp ;// use temp buffer instead
  mov  [ebp-4],eax
.init_out:
  ;//xor  edx,edx
  mov  [eax],edx   ;// Count_of_JSRuntime_found = 0
  mov  [eax+4],edx ;// Count_of_AbandonedChunks = 0
  add  eax,byte 8
  mov  [ebp-32],eax ;// BufWrite
  ;//
  call .pushed_EnumFunc
.EnumFunc: ;// eax=ebp_outer, edx=mbi
  push eax
  mov  eax,[edx+mbi_Address]
  push eax
  call IsAbandonedChunk
  pop  ecx ;// Address
  pop  edx ;// ebp_outer
  cmp  al,1
  je   .abandoned
  cmp  al,2
  jne  .eout
  mov  eax,[edx-4]
  inc  dword [eax] ;// Count_of_JSRuntime_found++
  jmp  short .eout
.abandoned:
  mov  eax,ecx
  call .AddAbandon
.eout:
  xor  eax,eax
  ret
  ;// sub-func:
.AddAbandon: ;// eax=Address, edx=ebp_outer
  mov  ecx,[edx-24] ;// sizeNeeded+=4
  add  ecx,byte 4
  mov  [edx-24],ecx
  cmp  ecx,[edx-12]
  jbe  .avail_ok
  ;// buffer overflow:
  xor  ecx,ecx
  mov  [edx-12],ecx  ;// sizeAvail:=0
  mov  [edx-28],ecx  ;// dwSkip:=0
  lea  ecx,[edx-48]
  mov  [edx-32],ecx  ;// BufWrite:=temp
.avail_ok:
  mov  ecx,[edx-4]
  inc  dword [ecx+4] ;// Count_of_AbandonedChunks++
  mov  ecx,[edx-32]  ;// BufWrite
  mov  [ecx],eax
  add  ecx,[edx-28]  ;// +=dwSkip
  mov  [edx-32],ecx
  ret
  ;//
.pushed_EnumFunc:
  pop  eax
  mov  edx,ebp
  call EnumVirtMap1M
  ;//
  ;// update out sizeNeeded
  mov  eax,[ebp-24]
  mov  edx,[ebp-8]
  mov  [edx],eax
  ;//
  ;// if(!Count_of_JSRuntime_found)return 0
  xor  eax,eax
  mov  edx,[ebp-4]
  cmp  dword [edx],eax
  je   .out
  ;// if(sizeNeeded<=sizeAvail)return sizeNeeded; else return -1
  mov  eax,[ebp-24]
  cmp  eax,[ebp-12]
  jbe  .out
  or   eax,byte -1
  ;//jmp short .out
  ;//
.out:
  mov  esp,ebp
  pop  ebp
  ret

;// function EnumVirtMap1M(Func,Arg): integer -- calls Func(Arg,mbi) for each 1Mb aligned block, until returns <>0...
EnumVirtMap1M:
  push ebp
  mov  ebp,esp
  push ebx
  push edi
  push esi
  ;//
  push byte 0    ;// [ebp-16] = Count of blocks found...
  push eax       ;// [ebp-20] = Func
  push edx       ;// [ebp-24] = Arg
  ;//
  sub  esp,byte mbi__sizeOf
  mov  edi,esp   ;// TMemoryBasicInformation
  ;// clear once:
  xor  eax,eax
  push byte mbi__sizeOf/4
  pop  ecx
  rep  stosd
  sub  edi,byte mbi__sizeOf
  ;//
  mov  ebx,100000h  ;// Address
  mov  esi,ebx      ;// min Step
  ;//
.cy1:
  ;// VirtualQuery(Address,mbi,size)
  push byte mbi__sizeOf
  push edi
  push ebx
  call VirtualQuery
  add  ebx,1000h ;//add  ebx,esi ;//add  ebx,100000h ;// minimum next address...
  test eax,eax
  jz   .vq_failed
  xor  eax,eax
  mov  eax,[edi+mbi_RegionSize]
  test eax,eax
  jle  .vq_failed
  ;//
  mov  edx,[edi+mbi_Address]
  add  eax,edx
  cmp  eax,ebx
  cmova ebx,eax
  ;//
  cmp  edx,[edi+mbi_AllocBase]
  jne  .different
  ;//
  cmp  esi,dword [edi+mbi_RegionSize] ;//100000h  ;// 1Mb
  jne  .different
  cmp  byte [edi+mbi_Protect],4 ;// Protect=4 PAGE_READWRITE
  jne  .different
  cmp  byte [edi+mbi_State+1],10h ;// State=0x1000 MEM_COMMIT
  jne  .different
  ;// this is (aligned??) 1Mb block, whole commited, read/write...
  mov  edx,edi
  mov  eax,[ebp-24]
  call dword [ebp-20]
  test eax,eax
  jnz  .break1
  ;//
.vq_failed:
.different:
  cmp  ebx,7FF00000h
  jb   .cy1
.break1:
  ;//jmp  short .out
.out:
  lea  esp,[ebp-12]
  pop  esi
  pop  edi
  pop  ebx
  pop  ebp
  ret

;// function IsAbandonedChunk(Address): Byte - returns 0 other, 1 abandoned, 2 valid JSRuntime Chunk
IsAbandonedChunk:
  push ebx
  push edi
  test eax,0fffffh
  jnz  .not_aligned
  ;//and  eax,-100000h
  add  eax,0fffc0h  ;// ^ChunkInfo
  mov  ebx,eax
  ;//
  add  eax,byte 3Ch ;// ^JSRuntime (last dword)
  call MmIsBadWrite
  test eax,eax
  jnz  .bad
  ;//
  mov  eax,[ebx+3Ch] ;// JSRuntime
  test eax,eax
  jle  .bad
  call Is_JSRuntime
  test al,al
  mov  al,2 ;// return 2 if valid JSRuntime found...
  jnz  .out
  ;//
  lea  edi,[ebx-32] ;// PerArenaBitmap should be FF...FF,0F
  cld
  push byte 31
  pop  ecx
  mov  al,0FFh
  repe scasb
  jne  .different
  mov  al,0Fh
  scasb
  jne  .different
  ;// here edi at fffc0
  ;// free-mark in padding at fffcc , no next,prev links etc...
  ;//push byte 3
  ;//pop  ecx
  xor  eax,eax
  add  ecx,byte 3
  repe scasd
  jne  .different
  mov  eax,[edi]
  scasd ;//add  edi,byte 4
  test eax,eax
  jz   .pad0_ok
  cmp  eax,'fre'
  jne  .different
.pad0_ok:
  ;//
  xor  eax,eax
  add  ecx,byte 6 ;// 6 more 00 dwords
  repe scasd
  jne  .different
  ;//
  mov  al,0FCh ;//mov  eax,0FCh
  scasd
  jne  .different
  xor  eax,eax
  scasd
  jnz  .different
  mov  al,2 ;//add  eax,byte 2
  scasd ;// ChunkInfo::location==2 ?
  jne  .different
  xor  eax,eax
  scasd
  jne  .different
  scasd
  jne  .different
  ;// this is abandoned chunk...
  mov  al,1
  jmp  short .out
.not_aligned:
.bad:
.different:
  xor  eax,eax
.out:
  pop  edi
  pop  ebx
  ret

;// function Is_JSRuntime(Address): Boolean byte -- detect JSRuntime structure...
;// in Palemoon 32, at +10h is pointer back to JSRuntime ... other checks are probably version dependent...
;// in Palemoon 32 JSRuntime is always at offset 0x118 in its page
;// dword at +0 is 0=Idle, 1..5=Busy
Is_JSRuntime:
  push eax
  ;//
  test al,3
  jnz  .bad
  ;//
  call MmIsBadWrite
  test al,al
  jnz  .bad
  pop  eax
  push eax
  mov  edx,eax
  add  eax,byte 10h
  ;// same page? (in Palemoon 32 always at offset 0x118 in page...)
  xor  edx,eax
  and  dx,-1000h
  jnz  .bad
  ;//
  mov  eax,[eax]
  cmp  eax,[esp]
  jne  .bad
  ;//
  cmp  dword [eax],byte 5
  setbe al
  jmp  short .out
.bad:
.done:
  setz al
.out:
  pop  edx
  ret

;//----------------------------------------------------------------
align 4
;// function ZeroTry(Proc,Param,ErrorHandler): Longint; stdcall;
;// silent dirty try-handler without unwinding any try-finally inside...
;// (used only during Startup to find kernel32.dll and during tests...)
;// in [esp+4]=[ebp+8]=Proc, [esp+8]=[ebp+12]=Param, [esp+12]=[ebp+16]=ErrorHandler
ZeroTry:
  push ebp
  mov  ebp,esp
  mov  eax,[ebp+12] ;//Param
  push eax
  push dword [ebp+16] ;// user ErrorHandler
  mov  edx,[ebp+8] ;// Proc
  push byte 0 ;// Result
  push ebx
  push esi
  push edi
  ;//
  call .2 ;//= push offset .continue
.continue:
  ;//??? lea  esp,[ebp-24]
  pop  edi
  pop  esi
  pop  ebx
  pop  eax
  ;//add  esp,byte 8 ;// skip internals ErrorHandler,Param
  pop  edx
  pop  edx
  ;//
  pop  ebp
  retn 12
  ;//
.2:
  call .3 ;//= push offset ExcHandler
.ExcHandler:
  ;//      { ->    [ESP+ 4] excPtr: PExceptionRecord       }
  ;//      {       [ESP+ 8] errPtr: PExcFrame              }
  ;//      {       [ESP+12] ctxPtr: Pointer                }
  ;//      {       [ESP+16] dspPtr: Pointer                }
  ;//      { <-    EAX return value - always one   }
  mov  edx,[esp+8]
  mov  ecx,[edx+1Ch] ;// ErrorHandler
  test ecx,ecx
  jz   .DoneExc
  ;//
  push ebp
  mov  ebp,esp
  ;//
  mov  eax,[esp+4+4] ;// EAX=^ExceptionRecord, EDX=^ExcFrame
  ;//
  push byte 0
  pop  dword [edx+1Ch]
  ;//
  call ecx
  ;//
  mov  esp,ebp
  pop  ebp
  ;//
  mov  edx,[esp+8]
  mov  [edx+18h],eax ;// change Result
  ;//
.DoneExc:
  mov  esp,[esp+8]
  xor  eax,eax
  db 64h,8Fh,00h ;//pop    fs:[eax]                         ; 00403428      64 8F 00
  pop  eax
  ret ;// -> pContinue... popa ...
  ;//
.3:
  xor  ecx,ecx
  db 64h,0FFh,31h ;//push   fs:[ecx]               ; 00403452      64 FF 31
  db 64h,89h,21h  ;//mov    fs:[ecx],esp           ; 00403455      64 89 21
  ;//
  push ebp
  mov  ebp,esp
  ;//
  push eax
  call edx
  ;//
  mov  esp,ebp
  pop  ebp
  ;//
  mov  [esp+18h],eax  ;// Result
  ;//
  xor  eax,eax
  db 64h,8Fh,00h ;//pop    fs:[eax]                         ; 00403467      64 8F 00
  add  esp,byte 8
  jmp short .continue

;//--------------------------------
;// function MmIsBadWrite(Ptr) - checks write access to one dword at [eax]
MmIsBadWrite:
  push ebp
  mov  ebp,esp
  push byte 1 ;// [ebp-4] = Result = bad unless changed...
  push ebx
  push esi
  push edi
  call .pushed_ehandler ;// = push offset ExcHandler
.ExcHandler:
  ;//      { ->    [ESP+ 4] excPtr: PExceptionRecord       }
  ;//      {       [ESP+ 8] errPtr: PExcFrame              }
  ;//      {       [ESP+12] ctxPtr: Pointer                }
  ;//      {       [ESP+16] dspPtr: Pointer                }
  ;//      { <-    EAX return value - always one   }
  mov  esp,[esp+8]
  jmp short .go_out
  ;//
.pushed_ehandler:
  xor  ecx,ecx
  db 64h,0FFh,31h ;//push   fs:[ecx]               ; 00403452      64 FF 31
  db 64h,89h,21h  ;//mov    fs:[ecx],esp           ; 00403455      64 89 21
  ;//
  or   [eax],ecx   ;// or 0, no change but check write access...
  mov  [ebp-4],ecx ;// return 0 if ok...
  ;//
.go_out:
  xor  eax,eax
  db 64h,8Fh,00h ;//pop    fs:[eax]                         ; 00403467      64 8F 00
  pop  edx ;// pop ExcHandler
  ;//jmp short .continue
.continue:
  pop  edi
  pop  esi
  pop  ebx
  pop  eax
  pop  ebp
  ret

;//-----------------------------------------------------------------
align 4

;// function GetProcAddress(hDll,ProcName): Pointer; stdcall;
;// in [ESP+4]=[EBP+8]=hDll , [ESP+8]=[EBP+12]=ProcName
GetProcAddress:
  push ebp
  mov  ebp,esp
  ;//
  push ebx
  push edi
  push esi
  ;//
  mov  ebx,[ebp+8]  ;// hDll
  test ebx,ebx
  jz   .out
  cmp  word [ebx],'MZ'
  jne  .fail
  mov  edx,[ebx+3Ch]
  add  edx,ebx
  cmp  dword [edx],4550h ;//'PE'
  jne  .fail
  mov  edx,[edx+78h] ;// RvaExports
  test edx,edx
  jz   .fail
  add  edx,ebx ;// TImageExportDirectory
  push edx ;// [esp] = ExportDir
  mov  ecx,[edx+18h] ;// NameCount
  test ecx,ecx
  jle  .fail
  mov  esi,[edx+20h]
  add  esi,ebx ;// NameTable
  xor  edi,edi
  cld
.ncy:
  lodsd
  test eax,eax
  jz   .nskip
  add  eax,ebx
  mov  edx,[ebp+12]
  push ecx
  call SameName
  pop  ecx
  test al,al
  jnz  .nfound
.nskip:
  inc  edi
  dec  ecx
  jnz  .ncy
  ;//jmp short .fail
.fail:
  xor  eax,eax
  jmp short .out
  ;//
.nfound:
  mov  eax,[esp] ;// ExportDir
  ;//
  mov  esi,[eax+24h] ;// rva NameOrdinalTable
  add  esi,ebx ;// NameOrdinalTable
  movzx edi,word [esi+2*edi]
  ;//
  mov  edx,[eax+1Ch] ;// rva FuncTable
  add  edx,ebx ;// FuncTable
  ;//x: add  edi,[eax+10h] ;// add ExportDir.OrdinalBase
  mov  edi,[edx+4*edi] ;// RVA
  add  edi,ebx ;// function
  cmp  edi,eax
  mov  eax,edi
  ;//ja   .forward
  jb   .out  ;// below ExportDir are normal functions...
  ;//mov  eax,edi
  ;//jmp short .out
.forward: ;// check if it is forwarded function 'NTDLL.RtlEnterCriticalSection'
  ;//mov  eax,edi
  call .isForward
  ;//jmp short .out
.out:
  lea  esp,[ebp-12]
  pop  esi
  pop  edi
  pop  ebx
  pop  ebp
  retn 8
  ;//
.isForward: ;// eax=func or name - check if it is forwarded function...
  mov  edx,eax
.cyfo1:
  mov cl,[eax]
  cmp cl,'.'
  je  .dotfo1
  sub cl,' '
  jl  .notfo1
  cmp cl,7Fh-' '
  jge .notfo1
  inc eax
  jmp .cyfo1
.notfo1:
  mov eax,edx ;// return original pointer...
  ret
.dotfo1:
  mov ecx,eax
  sub ecx,edx
  push ecx  ;// length of dllname without .dll
.cyfo2:
  inc eax
  mov cl,[eax]
  test cl,cl
  jz  .okfo2
  sub cl,' '
  jl  .notfo2
  cmp cl,7Fh-' '
  jl  .cyfo2
.notfo2:
  pop ecx
  jmp short .notfo1
.okfo2:
  mov  ecx,[esp]
  add  ecx,edx
  inc  ecx ;// start of funcname
  sub  eax,ecx ;// length of funcname
  jle  .notfo2
  ;// is forward name...
  ;// edx=DllName, ECX=FuncName, [esp]=[ebp+4]=length DllName
  push ebp
  mov  ebp,esp
  push edi
  push esi
  push edx ;// [ebp-12] = DllName - not terminated...
  push ecx ;// [ebp-16] = FuncName #0-terminated
  mov  eax,[ebp+4]
  add  eax,byte 4
  and  eax,byte -4
  sub  esp,eax
  mov  edi,esp
  push edi ;// param for GetModuleHandle
  mov  esi,edx
  mov  ecx,[ebp+4]
  rep  movsb
  xor  eax,eax
  stosb
  call GetModuleHandle
  ;//
  push dword [ebp-16]
  push eax
  call GetProcAddress
  ;//
  lea  esp,[ebp-8]
  pop  esi
  pop  edi
  pop  ebp
  pop  ecx
  ret
  ;//
  
;//---------------------------------------
;// used kernel32.dll functions:

GetModuleHandle: ;// GetModuleHandleA
  call get_Kernel
  jmp  dword [eax+k32_GetModuleHandle]
  ;//
VirtualFree:
  call get_Kernel
  jmp  dword [eax+k32_VirtualFree]
  ;//
VirtualQuery:
  call get_Kernel
  jmp  dword [eax+k32_VirtualQuery]
  ;//
    ;//CloseHandle:
    ;//  call get_Kernel
    ;//  jmp  dword [eax+k32_CloseHandle]
      ;//
    ;//CreateThread:
    ;//  call get_Kernel
    ;//  jmp  dword [eax+k32_CreateThread]
      ;//
Sleep:
  call get_Kernel
  jmp  dword [eax+k32_Sleep]
  ;//


;// function get_Kernel(Ptr): Pointer -- pass pointer to k32__sizeOf bytes...
get_Kernel:
  call getDataSeg
  add  eax,byte bss_Kernel32 ;// g_Kernel
  xor  edx,edx
  ;// note: first need GetModuleHandle before forwarded EnterCriticalSection:
  cmp  dword [eax+k32_GetModuleHandle],edx
  jz   init_Kernel32
  ret

;// function init_Kernel3(Ptr): Pointer -- pass pointer to k32__sizeOf bytes...
init_Kernel32: ;// in EAX=^g_Kernel
  call .here1
  .here1:
  pop  edx
  mov  edx,[edx+(imp__GetModuleHandle - .here1)]
  ;// jmp short init_Kernel2
;// function init_Kernel2(Ptr,GetModuleHandle): Pointer
init_Kernel2: ;// in EAX=^g_Kernel, EDX=GetModuleHandle if available...
  push eax
  call GetHKernel
  pop  edx
  ;//
  test eax,eax
  jnz  InitKernel
  ret
  ;//
InitKernel: ;// in EAX=HKernel32, EDX=^g_Kernel
  push edx ;// returns g_Kernel ...
  push edi
  push esi
  push ebx
  cld
  mov  edi,edx
  mov  esi,edx
  ;//
  stosd ;// g_Kernel[0] = EAX = HKernel32
  ;//
  push byte n_Kernel_Funcs
  pop  ecx
  ;//
  call .InitFuncs
.FuncNames:
  ;// n_Kernel_Funcs names ordered as in g_Kernel segment:
  dw .GetModuleHandle           - $  ;// must be first, required for forwarded functions...
  dw .VirtualFree               - $
  dw .VirtualQuery              - $
  dw .CloseHandle               - $
  dw .CreateThread              - $
  dw .Sleep                     - $
%ifdef Debug
  dw .FreeLibraryAndExitThread  - $
%endif ;Debug
.GetModuleHandle: db 'GetModuleHandleA',0
.VirtualFree: db 'VirtualFree',0
.VirtualQuery: db 'VirtualQuery',0
.CloseHandle: db 'CloseHandle',0
.CreateThread: db 'CreateThread',0
.Sleep: db 'Sleep',0
%ifdef Debug
  .FreeLibraryAndExitThread: db 'FreeLibraryAndExitThread',0
%endif ;Debug
  ;//
.InitFuncs:
  pop ebx ;// offset .FuncNames
  ;//
.fucy1:
  movzx eax,word [ebx]
  add   eax,ebx
  push  ecx
  ;//
  push  eax
  push  dword [esi]
  call  GetProcAddress
  stosd
  ;//
  pop  ecx
  add  ebx,byte 2
  dec  ecx
  jnz  .fucy1
  ;//
  pop  ebx
  pop  esi
  pop  edi
  pop  eax ;// return g_Kernel
  ret
  ;//

;// function GetHKernel(^g_Kernel,?GetModuleHandle)
GetHKernel: ;// in EAX=^g_Kernel, EDX=GetModuleHandle if available, returns [g_Kernel] possibly hKernel32 ...
  mov  ecx,eax
  mov  eax,[eax]
  test eax,eax
  jz   .find_Kernel
  ret
.find_Kernel:
  mov  eax,ecx
  test edx,edx
  jz   .find2
  ;//
  push eax
  ;// if we already have GetModuleHandle, use it:
  call .push_DllName
  db 'kernel32.dll',0
  .push_DllName:
  call edx
  test eax,eax
  jnz  .found2
  pop  eax
  ;//
.find2:
  push eax
  mov  eax,7F000000h  ;// start below this address...
.f2cy:
  sub  eax,10000h
  cmp  eax,esp
  jbe  .f2break
  ;//
  push eax
  call IsKernel32_dll
  test al,al
  pop  eax
  jz   .f2cy
.found2:
  mov  edx,eax
  pop  eax
  jmp short .found
.f2break:
  ;// not found ?!
  xor  edx,edx
  pop  eax
  ;//jmp short .found
  ;//
.found: ;// eax=^g_Kernel, edx=hKernel32, returns hKernel32
  mov  [eax],edx
  mov  eax,edx
  ret

;// function IsKernel32_dll(ImageBase)
IsKernel32_dll: ;// in EAX=hKernel32, out AL=1 if ok...
  push byte 0 ;// no ErrorHandler
  push eax    ;// Param
  call .chk2  ;// push offset .CheckProc
.CheckProc:
  call GetDllName
  test eax,eax
  jz   .none
  ;//call IsKernel32_Name
  ;// inline:
      ;// function IsKernel32_Name(Str)
      .IsKernel32_Name: ;// in EAX=^'KERNEL32.dll',0  out AL=1 if ok
        cmp  byte [eax+8],'.'
        jne  .diff
        mov  ecx,0DFDFDFDFh
        mov  edx,[eax]
        xor  edx,'KERN'
        and  edx,ecx
        jnz  .diff
        mov  edx,[eax+4]
        xor  edx,'EL32'
        and  edx,ecx
      .diff:
        setz al
        ;//ret
  ;//
.none:
  ret
.chk2:
  call ZeroTry
  ret

  
;// function GetDllName(ImageBase)
GetDllName: ;// in EAX=ImageBase, out EAX=0-terminated DllName 'KERNEL32.dll',0 or nil
  cmp  eax,10000h
  jb   __xar ;// .none
  cmp  word [eax],'MZ'
  jne  __xar ;// .none
  mov  edx,[eax+3Ch]
  add  edx,eax
  cmp  dword [edx],4550h ;// 'PE'
  jne  __xar ;// .none
  mov  edx,[edx+78h]
  test edx,edx
  jz   __xar ;// .none
  add  edx,eax ;// TImageExportDirectory
  mov  edx,[edx+12] ;// RvaName
  test edx,edx
  jz   __xar ;// .none
  add  eax,edx ;// ^Name
  ret
  ;//.none:
  ;//  xor  eax,eax
  ;//  ret

SameName: ;// eax=name1, edx=name2
  dec  eax
  dec  edx
.sn_next:
  inc  eax
  inc  edx
.sn_in:
  mov  cl,[eax]
  cmp  cl,[edx]
  jne .diffName
  test cl,cl
  jnz  .sn_next
  ;// same...
  mov  al,1
  ret
.diffName:
__xar:
  xor  eax,eax
  ret


;//-----------------------------------------------------------------

align 16,db 0
db 'Code by P.A.Semi',0

db 'V.02a',0

;//-----------------------------------------------
__end_Code equ $

;//size_Code equ __end_Code - start_Code

%ifdef DLL

;//align 1000h,db 0
TIMES (end_Code + 400h - ($-$$)) db 0

;//%include "MozGcGc_exports.inc"

%endif ;DLL
